// These define special actions that take more than one keystroke
#define NO_ACTION 0x00
#define ACTION_UNDO -1
#define ACTION_HIGH_CONTRAST -3
#define ACT_TRACK_PREV -4
#define ACT_GRID_BIGGER -5
#define ACT_MEASURE_SETGRID -6
#define ACT_SAVE -8
#define ACT_UNDO_REDO -9
#define ACT_RELATIVE -10
#define ACT_ESC_TAB -11
#define ACT_ENTER_DELETE -12

// ugly global state variables because I need to get this done.
// The modifier key is currently physically pressed by user
// used for alternate actions
bool ctrlDown = false;
bool shiftDown = false;

// This maps matrix key positions to flat unique control IDs
const byte keyMatrix[4][4] = {
    {0, 1, 2, 3},
    {4, 5, 6, 7},
    {8, 9, 10, 11},
    {12, 13, 14, 15}
  };

// yucky. This shouldn't be required, but here we are.
// defines which keys are holdable (corresponds to ctrl, shift, up, down, left, right)
const bool holdableKeys[4][4] = {
    {0, 0, 0, 0},
    {0, 0, 1, 0},
    {1, 1, 1, 1},
    {1, 1, 0, 0}
  };

  // This maps encoder buttons to flat unique control IDs
const byte encoderButtons[3] = {16, 17, 18};
const int numEncoders = (sizeof(encoderButtons) / sizeof(*encoderButtons));

// This array is indexed by the unique control ID, mapping a button to an action
// [0-15] matrix keypad
// [16-18] encoder buttons
const int switchAssignments[] = {
  KEY_X, KEY_D, KEY_V, ACT_MEASURE_SETGRID,
  ACTION_HIGH_CONTRAST, KEY_R, KEY_UP, KEY_M,
  MODIFIERKEY_SHIFT, KEY_LEFT, KEY_DOWN, KEY_RIGHT,
  MODIFIERKEY_CTRL, ACT_UNDO_REDO, ACT_ENTER_DELETE, ACT_ESC_TAB,
  ACT_SAVE, KEY_MEDIA_NEXT_TRACK, KEY_MEDIA_PLAY_PAUSE
};

// Count number of switches assigned
const int numSwitches = (sizeof(switchAssignments) / sizeof(*switchAssignments));

// Encoders
// [Left, Middle, Right][Anti-clockwise, Clockwise]
const int knobAssignments[3][2] = { {ACT_TRACK_PREV, KEY_W},
                                    {ACT_GRID_BIGGER, KEY_N},
                                    {KEY_MEDIA_VOLUME_DEC, KEY_MEDIA_VOLUME_INC}};

void releaseAllKeys() {
  Keyboard.set_modifier(0);
  Keyboard.set_key1(0);
  Keyboard.send_now();
}

void performAction(int action) {
  // Actions takke more than one keystroke OR
  // begin a hold event for a key if it is defined here
  switch (action) {
    case NO_ACTION:
      break;

    case ACTION_HIGH_CONTRAST: // Ctrl + H
      Keyboard.press(MODIFIERKEY_CTRL);
      Keyboard.press(KEY_H);
      Keyboard.release(KEY_H);
      Keyboard.release(MODIFIERKEY_CTRL);
      break;

    case ACT_TRACK_PREV:
      Keyboard.press(MODIFIERKEY_SHIFT);
      Keyboard.press(KEY_W);
      Keyboard.release(KEY_W);
      Keyboard.release(MODIFIERKEY_SHIFT);
      break;

    case ACT_GRID_BIGGER:
      Keyboard.press(MODIFIERKEY_SHIFT);
      Keyboard.press(KEY_N);
      Keyboard.release(KEY_N);
      Keyboard.release(MODIFIERKEY_SHIFT);
      break;

    case ACT_MEASURE_SETGRID:
      if (ctrlDown) {
        Keyboard.release(MODIFIERKEY_CTRL);
        Keyboard.press(KEY_S);
        Keyboard.release(KEY_S);
        Keyboard.press(MODIFIERKEY_CTRL);
      } else {
        Keyboard.set_modifier(MODIFIERKEY_CTRL | MODIFIERKEY_SHIFT);
        Keyboard.press(KEY_M);
        // Keyboard.send_now();
        releaseAllKeys();
      }
      break;

    case ACT_SAVE:
      Keyboard.press(MODIFIERKEY_CTRL);
      Keyboard.press(KEY_S);
      releaseAllKeys();
      break;

    case MODIFIERKEY_CTRL:
      Keyboard.set_modifier(MODIFIERKEY_CTRL);
      Serial.println("Ctrl down");
      ctrlDown = true;
      break;

    case MODIFIERKEY_SHIFT:
      Keyboard.set_modifier(MODIFIERKEY_SHIFT);
      Serial.println("Shift down");
      shiftDown = true;
      break;

    case KEY_UP:
      Keyboard.press(KEY_UP);
      break;

    case KEY_DOWN:
      Keyboard.press(KEY_DOWN);
      break;

    case KEY_LEFT:
      Keyboard.press(KEY_LEFT);
      break;

    case KEY_RIGHT:
      Keyboard.press(KEY_RIGHT);
      break;

    case KEY_M:
      // M = Move
      // Ctrl+M = Move exactly
      // Shift+M => Ctrl+R Position relative to
      if (shiftDown) {
        Keyboard.release(MODIFIERKEY_SHIFT);
        Keyboard.set_modifier(MODIFIERKEY_CTRL);
        Keyboard.press(KEY_R); // Ctrl+R: move relative to
        Keyboard.release(KEY_R); //send ctrl+ only
        Keyboard.release(MODIFIERKEY_CTRL);
        Keyboard.set_modifier(MODIFIERKEY_SHIFT); // restore to how user is physically pressing
      } else { // regular move
        Keyboard.press(action);
        Keyboard.release(action);
      }
      break;

    case KEY_X:
      // X = add track
      // Ctrl+X => "/" change track posture
      if (ctrlDown) {
        Keyboard.release(MODIFIERKEY_CTRL);
        Keyboard.press(KEY_SLASH);
        Keyboard.release(KEY_SLASH);
        Keyboard.set_modifier(MODIFIERKEY_CTRL); // restore to how user is physically pressing
      } else { // regular add track
        Keyboard.press(action);
        Keyboard.release(action);
      }
      break;

      case ACT_UNDO_REDO: // holdable
        if (ctrlDown) { // redo
          Keyboard.set_modifier(MODIFIERKEY_CTRL);
          Keyboard.press(KEY_Y);
        } else { // undo
          Keyboard.set_modifier(MODIFIERKEY_CTRL);
          Keyboard.press(KEY_Z);
          // releaseAllKeys();
        }
        break;

        case KEY_D:
          if (ctrlDown) { // drag, any angle
            Keyboard.release(MODIFIERKEY_CTRL);
            Keyboard.press(KEY_G);
            Keyboard.release(KEY_G);
            Keyboard.press(MODIFIERKEY_CTRL);
          } else { // undo
            Keyboard.press(action);
            Keyboard.release(action);
          }
          break;

        case ACT_ESC_TAB: // Esc or Ctrl+Tab
          if (ctrlDown) { // drag, any angle
            Keyboard.release(MODIFIERKEY_CTRL);
            Keyboard.press(MODIFIERKEY_ALT);
            Keyboard.press(KEY_TAB);
            Keyboard.release(KEY_TAB);
            Keyboard.release(MODIFIERKEY_ALT);
            Keyboard.press(MODIFIERKEY_CTRL);
          } else {
            Keyboard.press(KEY_ESC);
            Keyboard.release(KEY_ESC);
          }
          break;

        case ACT_ENTER_DELETE:
          if (ctrlDown) {
            Keyboard.release(MODIFIERKEY_CTRL);
            Keyboard.press(KEY_DELETE);
            Keyboard.release(KEY_DELETE);
            Keyboard.press(MODIFIERKEY_CTRL);
          } else {
            Keyboard.press(KEY_ENTER);
            Keyboard.release(KEY_ENTER);
          }
          break;

    default: // regular keys with no alt functions
      Keyboard.press(action);
      Keyboard.release(action);
  }
}
